/*
 * me26.c -- Driver for the Meilhaus me-26 A/D converter board
 * me26.c -- Written by Xavier Calbet.
 * 
 *    Version 0.5
 *
 * This software is released under the GPL
 *
 *
 *
 * 
 *********/

#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif


#include <linux/module.h>

#include <linux/sched.h>
#include <linux/kernel.h> /* printk() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/tqueue.h>
#include <linux/delay.h>
#include <linux/time.h>

#include <asm/io.h>
#include <asm/segment.h>

#include "me26.h"

/* Number of cards on the same PC, if you want to add
more, besides changing this number you must change
the arrays me26_base_d and me26_irq_t and add more me26_basex
and me26_irqx variables */
int me26_nr_cards = ME26_NR_CARDS;

int me26_major = ME26_MAJOR;

/* Card number */
int me26_cn;

/* Devices or cards */
Me26_Dev *me26_devices;

/* Variables of base addresses to be used externally when
installing the module */
int me26_base  = 0;
int me26_base2 = 0;
int me26_base3 = 0;
int me26_base4 = 0;
/* Array of base addresses for each card */
int me26_base_t[ME26_NR_CARDS + 1];

/* Variables of IRQs to be used externally when installing
the module */
int me26_irq  = -1;
int me26_irq2 = -1;
int me26_irq3 = -1;
int me26_irq4 = -1;
/* Array of IRQs for each card */
volatile int me26_irq_t[ME26_NR_CARDS];
/* Inverse IRQ array */
int me26_card_t[16];

/* Wait queues */
struct wait_queue *me26_queue[ME26_NR_CARDS];





#if LINUX_VERSION_CODE < VERSION_CODE(1,3,70)
void me26_probing(int irq, struct pt_regs *regs)
#else
void me26_probing(int irq, void *dev_id, struct pt_regs *regs)
#endif
{
    if (me26_irq_t[me26_cn] == 0) me26_irq_t[me26_cn] = irq;    /* found */
    if (me26_irq_t[me26_cn] != irq) me26_irq_t[me26_cn] = -irq;/* ambiguous */
}

void me26_selfprobe(int lcn)
{
    int trials[] = {3, 4, 5, 6, 7, 0};
    int tried[]  = {0, 0, 0, 0, 0, 0};
    int i, count = 0;

    /* If we give the IRQ we use only that one */
    if (me26_irq_t[lcn] >= 0) {
      trials[0] = me26_irq_t[lcn];
      trials[1] = 0 ;
    }


     /*
      * install the probing handler for all possible lines. Remember
      * the result (0 for success, or -EBUSY) in order to only free
      * what has been acquired
      */
    for (i=0; trials[i]; i++)
        tried[i] = request_irq(trials[i], me26_probing,
                               SA_INTERRUPT, "me26 probe", NULL);

    do {
        me26_irq_t[lcn] = 0; /* none got, yet */
        outb_p(CHANNEL06 | SOFT_TRIGGER | GO | ENABLE_INTERRUPT, 
	     me26_base_t[lcn] + AD_CONTROL); /* enable interrupt and trigger*/
	/* We wait 50 microseconds */
	udelay(50);
	outb_p(CHANNEL06 | SOFT_TRIGGER, 
	       me26_base_t[lcn] + AD_CONTROL); /* disable interrupt */

        /* the value has been set by the handler */
        if (me26_irq_t[lcn] == 0) { /* none of them? */
            printk(KERN_INFO "me26[%d]: no irq reported by probe\n",lcn);
        }
        /*
         * If more than one line has been activated, the result is
         * negative. We should service the interrupt
         * and loop over again. Do it at most 5 times
         */
    } while (me26_irq_t[lcn] <=0 && count++ < 5);

    /* end of loop, uninstall the handler */
    for (i=0; trials[i]; i++)
        if (tried[i] == 0)
            free_irq(trials[i], NULL);

    if (me26_irq_t[lcn] <= 0)
        printk(KERN_INFO 
	       "me26[%d]: probe failed %i times, giving up\n",lcn, count);
}



/*
 * The ioctl() implementation
 */

int me26_ioctl (struct inode *inode, struct file *filp,
                 unsigned int cmd, unsigned long arg)
{

    unsigned int card = CARD(inode->i_rdev);
    unsigned char i_channel = CHANNEL(inode->i_rdev);


    Me26_Dev *dev=filp->private_data;
    unsigned char * timer_byte;
    unsigned char channel,data;

    
    /* We set the contents of channel variable to a useful value */ 
    channel=i_channel;
    channel<<=4;

    /*
     * extract the type and number bitfields, and don't decode
     * wrong cmds: return EINVAL
     */
    if (_IOC_TYPE(cmd) != ME26_IOC_MAGIC) return -EINVAL;
    if (_IOC_NR(cmd) > ME26_IOC_MAXNR) return -EINVAL;

    PDEBUG("me26_ioctl[%d]: channel 0x%x, i_channel 0x%x, arg 0x%lx\n",
	   card,channel,i_channel,arg);

    switch(cmd) {

      case ME26_IOCEXTTIMER:
	PDEBUG("me26_ioctl[%d]: option IOCEXTTIMER\n",card);
	PDEBUG("me26_ioctl[%d]: channel 0x%x, i_channel 0x%x\n",
	       card,channel,i_channel);
        dev->trigger=0;
        break;
        
      case ME26_IOCINTTIMER:
	PDEBUG("me26_ioctl[%d]: option IOCINTTIMER\n",card);
	PDEBUG("me26_ioctl[%d]: channel 0x%x, i_channel 0x%x\n",
	       card,channel,i_channel);
        dev->trigger=1;
	timer_byte=(unsigned char *)&arg;
	/* Set counter0 */
        outb_p(SET_COUNTER_0, 
	     me26_base_t[card] + TIMER_CONTROL);
	/* LSB of timer 0 */
	PDEBUG("me26_ioctl[%d]: writing to COUNTER_0, LSB 0x%x\n",
	       card,*(timer_byte+3));
	outb_p( *(timer_byte+3),
	     me26_base_t[card] + COUNTER_0 );
	/* MSB of timer 0 */
	PDEBUG("me26_ioctl[%d]: writing to COUNTER_0, MSB 0x%x\n",
	       card,*(timer_byte+2));
	outb_p( *(timer_byte+2),
	     me26_base_t[card] + COUNTER_0 );
	/* Set counter 1 */
	outb_p( SET_COUNTER_1,
	     me26_base_t[card] + TIMER_CONTROL);
	/* LSB of timer 1 */
	PDEBUG("me26_ioctl[%d]: writing to COUNTER_1, LSB 0x%x\n",
	       card,*(timer_byte+1));
	outb_p( *(timer_byte+1),
	     me26_base_t[card] + COUNTER_1 );
	/* MSB of timer 1 */
	PDEBUG("me26_ioctl[%d]: writing to COUNTER_1, MSB 0x%x\n",
	       card,*(timer_byte+0));
	outb_p( *(timer_byte+0),
	     me26_base_t[card] + COUNTER_1 );
	/* Enable internal timer */
	if (dev->interrupt == 0) {
	  outb_p(channel,
		 me26_base_t[card] + AD_CONTROL);
	} else {
	  outb_p(channel | ENABLE_INTERRUPT,
		 me26_base_t[card] + AD_CONTROL);
	}	  
		
        break;

      case ME26_IOCSOFTIMER:
	PDEBUG("me26_ioctl[%d]: option IOCSOFTIMER\n",card);
	PDEBUG("me26_ioctl[%d]: channel 0x%x, i_channel 0x%x\n",
	       card,channel,i_channel);
        dev->trigger=2;
        break;

      case ME26_IOCUSEINTERRUPT:
	PDEBUG("me26_ioctl[%d]: option IOCUSEINTERRUPT\n",card);
	PDEBUG("me26_ioctl[%d]: channel 0x%x, i_channel 0x%x\n",
	       card,channel,i_channel);
        dev->interrupt=1;
	if (dev->trigger <= 1 ) {
	  /* External timer or internal timer*/
	  outb_p(channel | ENABLE_INTERRUPT ,
		 me26_base_t[card] + AD_CONTROL);
	} else {
	  /* Soft trigger */
	  /* outb_p(channel | SOFT_TRIGGER, 
	   me26_base_t[card] + AD_CONTROL); */
	  data=inb_p(me26_base_t[card] + AD_CONTROL);
	  PDEBUG("me26_ioctl[%d]: data 0x%x, channel 0x%x, ad 0x%x\n",
		 card,data,channel,
		 channel | SOFT_TRIGGER | ENABLE_INTERRUPT);
	  outb_p(channel | SOFT_TRIGGER | ENABLE_INTERRUPT ,
		 me26_base_t[card] + AD_CONTROL);
	  
	}

	udelay(50);
	/* Clearing the reading buffer. this has to be done becuase
	   when a transition from
	   ENABLE_INTERRUPT to NOINTERRUPT happens
	   an interrupt is started */
	  PDEBUG("me26_ioctl[%d]: clearing buffer: channel 0x%x\n",card,channel);
	dev->tail[i_channel]=dev->head[i_channel];
	dev->timtail=dev->timhead;
	/* Restoring initial channel  if changed by me26_interrupt */
	if (dev->trigger <= 1) {
	  /* External timer or internal timer*/
	  outb_p(channel | ENABLE_INTERRUPT ,
		 me26_base_t[card] + AD_CONTROL);
	} else {
	  /* Soft trigger */
	  outb_p(channel | SOFT_TRIGGER | ENABLE_INTERRUPT ,
		 me26_base_t[card] + AD_CONTROL);
	}
	  
        break;

      case ME26_IOCNOINTERRUPT:
	PDEBUG("me26_ioctl[%d]: option IOCNOINTERRUPT\n",card);
	PDEBUG("me26_ioctl[%d]: channel 0x%x, i_channel 0x%x\n",
	       card,channel,i_channel);
	dev->interrupt=0;
	if (dev->trigger <= 1 ) {
	  /* External timer or internal timer */
	  outb_p(channel ,
		 me26_base_t[card] + AD_CONTROL);
	} else {
	  /* Soft trigger */
	  outb_p(channel | SOFT_TRIGGER ,
		 me26_base_t[card] + AD_CONTROL);
	}	  
	  
	break;


      default:  /* redundant, as cmd was checked against MAXNR */
        return -EINVAL;
    }

    PDEBUG("me26_ioctl[%d]: trigger %d\n",card,dev->trigger);
    PDEBUG("me26_ioctl[%d]: interrupt %d\n",card,dev->interrupt);
    return 0;

}


/* The reading routine */

read_write_t me26_read (struct inode *inode, struct file *filp,
                char *buf, count_t count)
{
    unsigned int card = CARD(inode->i_rdev);
    unsigned char i_channel = CHANNEL(inode->i_rdev);
    
    unsigned char *kbuf, *ptr;

    Me26_Dev *dev=filp->private_data;
    unsigned char channel;
    unsigned char mask;

    int retval;
    int count0;

    struct timeval tv;
    int * secs;

    /* Reading time buffer */
    if ( i_channel == 16 ) {
      /* Count must be a multiple of 8 */
      count >>= 3;
      count <<= 3;
      PDEBUG("me26_read[%d]: channel 0x%x, count 0x%x\n",
	   card,i_channel,count);

      if (dev->timhead != dev->timtail) {
	/* count0 is the number of readable data bytes */
	count0 = dev->timhead - dev->timtail;
	if (count0 < 0)  /* wrapped */
	  count0 = dev->timbuf + PAGE_SIZE 
	    - dev->timtail;
	if (count0 < count) count = count0;
	PDEBUG("me26_read[%d]: Reading timer 0x%x, count 0x%x\n",
	   card,i_channel,count);
	memcpy_tofs(buf, (char *)dev->timtail, count);
	dev->timtail += count;
	if (dev->timtail == dev->timbuf + PAGE_SIZE)
	  dev->timtail = dev->timbuf;
	return count;
      } else {
	PDEBUG("me26_read[%d]: NOT reading timer 0x%x, count 0x%x\n",
	   card,i_channel,count);
	return 0;
      }
    }
      


    /* Checking if count is even */
    if (count & 0x01)
      count--;

    retval=count;

    if (dev->trigger == 0) {
      mask=0x1F;
    } else {
      mask=0x0F;
    }

    /* We set the contents of channel variable to a useful value */ 
    channel=i_channel<<4;
    PDEBUG("me26_read[%d]: channel 0x%x, i_channel 0x%x, mask 0x%x\n",
	   card,channel,i_channel,mask);


    if (dev->interrupt == 0) {
      /* NOT using interrupts */
      kbuf=kmalloc(count,GFP_KERNEL);
      if (!kbuf) return -ENOMEM;
      ptr=kbuf;
      
      while (count) {
	if (dev->trigger == 2) {
	  /* Soft trigger */
	  outb_p(channel | SOFT_TRIGGER | GO, 
		 me26_base_t[card] + AD_CONTROL);
	  /* 40 microseconds waiting */
	  udelay(40);

	  /* Select channel for reading */
	  outb_p(channel | SOFT_TRIGGER, 
		 me26_base_t[card] + AD_CONTROL);
	  *ptr=inb_p(me26_base_t[card] + AD_LSB);
	  ptr++;
	  *ptr=( inb_p(me26_base_t[card] + AD_MSB) &
		     mask );
	  ptr++;

	  if (dev->timbuf != 0 ) {
	    /* Reading time */
	    do_gettimeofday(&tv);
	    PDEBUG("me26_interrupt[%d]: end of reading data: channel 0x%x, i_channel 0x%x, times %d\n",card,channel,i_channel,tv.tv_sec);
	    PDEBUG("me26_interrupt[%d]: end of reading data: channel 0x%x, i_channel 0x%x, usec %d\n",card,channel,i_channel,tv.tv_usec);
	    secs=(int *)dev->timhead; 
	    *secs=tv.tv_sec;
	    secs++;
	    *secs=tv.tv_usec;
	    secs++;
	    dev->timhead=(unsigned long)secs;
	    if (dev->timhead == dev->timbuf + PAGE_SIZE)
	      dev->timhead = dev->timbuf; /* wrap */
	  }
	} else {
	
	  if (dev->trigger != 0 )
	    /* 40 microseconds waiting */
	    udelay(40);

	  *ptr=inb_p(me26_base_t[card] + AD_LSB);
	  ptr++;
	  *ptr=( inb_p(me26_base_t[card] + AD_MSB) &
		     mask );
	  ptr++;

	  if (dev->timbuf != 0 ) {
	    /* Reading time */
	    do_gettimeofday(&tv);
	    PDEBUG("me26_interrupt[%d]: end of reading data: channel 0x%x, i_channel 0x%x, times %d\n",card,channel,i_channel,tv.tv_sec);
	    PDEBUG("me26_interrupt[%d]: end of reading data: channel 0x%x, i_channel 0x%x, usec %d\n",card,channel,i_channel,tv.tv_usec);
	    secs=(int *)dev->timhead; 
	    *secs=tv.tv_sec;
	    secs++;
	    *secs=tv.tv_usec;
	    secs++;
	    dev->timhead=(unsigned long)secs;
	    if (dev->timhead == dev->timbuf + PAGE_SIZE)
	      dev->timhead = dev->timbuf; /* wrap */
	  }

	}
	count-=2;

      }

      memcpy_tofs(buf,kbuf,retval);
      kfree(kbuf);
      return retval;

    } else {
      /* Using interrupts */
      if (dev->trigger == 2) {
	/* Soft trigger */
	PDEBUG("me26_read[%d]: soft trigger\n",card);
	outb_p(channel | ENABLE_INTERRUPT | SOFT_TRIGGER, 
	       me26_base_t[card] + AD_CONTROL);
	outb_p(channel | ENABLE_INTERRUPT | SOFT_TRIGGER | GO, 
	       me26_base_t[card] + AD_CONTROL);
      }
      
      while (dev->head[i_channel] == dev->tail[i_channel]) {
	PDEBUG("me26_read[%d]: before sleeping channel 0x%x, i_channel 0x%x, mask 0x%x\n",card,channel,i_channel,mask);
	interruptible_sleep_on(&me26_queue[card]);
	PDEBUG("me26_read[%d]: waking up channel 0x%x, i_channel 0x%x, mask 0x%x\n",card,channel,i_channel,mask);
	if (current->signal & ~current->blocked) /* a signal arrived */
	  return -ERESTARTSYS; /* tell the fs layer to handle it */
	/* else, loop */
      }
      PDEBUG("me26_read[%d]: tranfering data to user 0x%x, i_channel 0x%x, mask 0x%x\n",card,channel,i_channel,mask);
     /* count0 is the number of readable data bytes */
      count0 = dev->head[i_channel] - dev->tail[i_channel];
      if (count0 < 0) /* wrapped */
	count0 = dev->buffer[i_channel] + PAGE_SIZE 
	  - dev->tail[i_channel];
      if (count0 < count) count = count0;
      memcpy_tofs(buf, (char *)dev->tail[i_channel], count);
      dev->tail[i_channel] += count;
      if (dev->tail[i_channel] == dev->buffer[i_channel] + PAGE_SIZE)
	dev->tail[i_channel] = dev->buffer[i_channel];
      return count;
    }
}


/*
 * The interrupt handler has a different prototype whether it
 * is compiled with kernels older or newer than 1.3.70
 */

#if LINUX_VERSION_CODE < VERSION_CODE(1,3,70)
void me26_interrupt(int irq, struct pt_regs *regs)
#else
void me26_interrupt(int irq, void *dev_id, struct pt_regs *regs)
#endif
{
    Me26_Dev * dev;

    int card;
    unsigned char mask;
    unsigned char i_channel;
    unsigned char channel;
    unsigned char * ptr;
    unsigned char data;
    struct timeval tv;
    int * secs;

    /* Finding card number */
    card=me26_card_t[irq];
    
    dev=&me26_devices[card];

    /* Finding channel number */
    channel=inb_p(me26_base_t[card] + AD_CONTROL);
    i_channel=channel>>4;


    if (dev->trigger == 0) {
      mask=0x1F;
    } else {
      mask=0x0F;
    }


    PDEBUG("me26_interrupt[%d]: channel 0x%x, i_channel 0x%x, mask 0x%x\n",
	   card,channel,i_channel,mask);

    if (dev->trigger == 2) {
      /* Prepare to read */
      outb_p(channel | ENABLE_INTERRUPT | SOFT_TRIGGER, 
	     me26_base_t[card] + AD_CONTROL);
    }

    if (dev->buffer[i_channel] != 0 ) {
      PDEBUG("me26_interrupt[%d]: reading data: channel 0x%x, i_channel 0x%x, mask 0x%x\n",card,channel,i_channel,mask);
      ptr=(unsigned char *)dev->head[i_channel];
      *ptr=inb_p(me26_base_t[card] + AD_LSB);
      ptr++;
      *ptr=( inb_p(me26_base_t[card] + AD_MSB) &
	     mask );
      ptr++;
      dev->head[i_channel]=(unsigned long)ptr;
      
     
      if (dev->head[i_channel] == dev->buffer[i_channel] + PAGE_SIZE)
        dev->head[i_channel] = dev->buffer[i_channel]; /* wrap */
      PDEBUG("me26_interrupt[%d]: end of reading data: channel 0x%x, i_channel 0x%x, mask 0x%x\n",card,channel,i_channel,mask);

      if (dev->timbuf != 0) {
	/* Reading time */
	do_gettimeofday(&tv);
	PDEBUG("me26_interrupt[%d]: end of reading data: channel 0x%x, i_channel 0x%x, times %d\n",card,channel,i_channel,tv.tv_sec);
	PDEBUG("me26_interrupt[%d]: end of reading data: channel 0x%x, i_channel 0x%x, usec %d\n",card,channel,i_channel,tv.tv_usec);
	secs=(int *)dev->timhead; 
	*secs=tv.tv_sec;
	secs++;
	*secs=tv.tv_usec;
	secs++;
	dev->timhead=(unsigned long)secs;
	if (dev->timhead == dev->timbuf + PAGE_SIZE)
	  dev->timhead = dev->timbuf; /* wrap */
      }
    } else {
      /* Reading to not block interrupts */
      data=inb_p(me26_base_t[card] + AD_LSB);
      data=inb_p(me26_base_t[card] + AD_MSB);
    }     
    
    
    /* If we are using internal or external triggering, we
       add one to the channel number to read the next channel
       number on the next trigger */
    if (dev->trigger != 2) {
      i_channel++;
      if (i_channel > 15) i_channel=0;
      channel=i_channel<<4;
      PDEBUG("me26_interrupt[%d]: change: channel 0x%x, i_channel 0x%x, mask 0x%x, writing 0x%x\n",card,channel,i_channel,mask,channel | ENABLE_INTERRUPT);
      
      outb_p(channel | ENABLE_INTERRUPT,
	     me26_base_t[card] + AD_CONTROL);
    }   

    PDEBUG("me26_interrupt[%d]: waking up\n",card);
    wake_up_interruptible(&me26_queue[card]); /* awake any reading process */
}

/*
 * Open and close
 */

int me26_open (struct inode *inode, struct file *filp)
{
    int card = CARD(inode->i_rdev);
    unsigned char i_channel = CHANNEL(inode->i_rdev);
    Me26_Dev *dev; /* device information */
    int result_irq;
    unsigned char channel;
    unsigned char data;

    /* manage errors first */
    
    if (card > me26_nr_cards) return -ENODEV;
    dev=&me26_devices[card];

    /* Reserving time buffer */
    if (i_channel == 16) {
      if (dev->timsin) return -EBUSY;
      dev->timsin=1;
      dev->timbuf=__get_free_page(GFP_KERNEL);
      dev->timhead=dev->timbuf;
      dev->timtail=dev->timbuf;
      PDEBUG("me26_open[%d]: channel 0x%x, buffer 0x%lx\n",
	     card,i_channel, dev->timbuf);
      PDEBUG("me26_open[%d]: channel 0x%x, head 0x%lx\n",
	     card,i_channel, dev->timhead);
      PDEBUG("me26_open[%d]: channel 0x%x, tail 0x%lx\n",
	     card,i_channel, dev->timtail);
      filp->private_data=dev;
      MOD_INC_USE_COUNT;
      return 0;          /* success */
    }

    if (i_channel > 15) return -ENODEV;
    if (dev->single[i_channel]) return -EBUSY;
    dev->single[i_channel]=1;

    channel=i_channel<<4;

    PDEBUG("me26_open[%d]: channel 0x%x, i_channel 0x%x\n",
	   card,channel, i_channel);


    /* Default configuration, can be changed with ioctl */
    dev->trigger=2; /* Software trigger A/D */
    dev->interrupt=0; /* Do not use interrupts */
    outb_p(channel | SOFT_TRIGGER, 
       me26_base_t[card] + AD_CONTROL); /* Telling the board */
    data=inb_p(me26_base_t[card] + AD_CONTROL);
    PDEBUG("me26_open[%d]: channel 0x%x, i_channel 0x%x AD-CONTROL 0x%x\n",
	   card,channel, i_channel,data);
    /* Setting things up for reading buffer */
    dev->buffer[i_channel]=__get_free_page(GFP_KERNEL);
    dev->head[i_channel]=dev->buffer[i_channel];
    dev->tail[i_channel]=dev->buffer[i_channel];


    /* Reserving IRQ (only once for the card) */
    if (dev->req_int == 0) {
      PDEBUG("me26_open[%d]: requesting interrupt %d: channel 0x%x, i_channel 0x%x\n",card,me26_irq_t[card],channel, i_channel);
      result_irq = request_irq(me26_irq_t[card], me26_interrupt,
			       SA_INTERRUPT, "me26", NULL);
      if (result_irq) {
	printk(KERN_INFO "me26[%d]; can't get assigned irq %d\n",
	       card, me26_irq_t[card]);
	return result_irq;
      }
      dev->req_int=1;
    }

    /* and use filp->private_data to point to the device data */
    filp->private_data = dev;

    MOD_INC_USE_COUNT;
    return 0;          /* success */
}

void me26_release (struct inode *inode, struct file *filp)
{
    int card = CARD(inode->i_rdev);
    unsigned char i_channel = CHANNEL(inode->i_rdev);
    Me26_Dev *dev=filp->private_data; /* device information */
    unsigned char i;
    unsigned char total;
    unsigned char channel;

    channel=i_channel<<4;

    PDEBUG("me26_release[%d]: i_channel 0x%x channel 0x%x\n",
	   card,i_channel,channel);

    if (i_channel == 16) {
      dev->timsin=0;
      free_page(dev->timbuf);
      dev->timbuf=0;
      MOD_DEC_USE_COUNT;
      return;
    }
    

    free_page(dev->buffer[i_channel]);
    dev->buffer[i_channel]=0;
    dev->single[i_channel]=0;


    /* Checking if all channels are free or not */
    total=0;
    for (i=0;i<16;i++) {
      if (dev->buffer[i] != 0) {
	total=1;
	break;
      }
    }

    /* If all channels are free release interrupt */
    if (total == 0 ) {
      /* Returning to default configuration with no interrupts */
      dev->trigger=2; /* Software trigger A/D */
      dev->interrupt=0; /* Do not use interrupts */
      outb_p(channel | SOFT_TRIGGER, 
      	     me26_base_t[card] + AD_CONTROL); /* Telling the board */
      PDEBUG("me26_release[%d]: releasing interrupt: i_channel 0x%x\n",
	     card,i_channel);
      free_irq(me26_irq_t[card],NULL);
      dev->req_int=0;
    }

    MOD_DEC_USE_COUNT;
}



struct file_operations me26_fops = {
    NULL,          /* me26_lseek */
    me26_read,
    NULL,          /* me26_write */
    NULL,          /* me26_readdir */
    NULL,          /* me26_select */
    me26_ioctl,          /* me26_ioctl */
    NULL,          /* me26_mmap */
    me26_open,
    me26_release,
    NULL,          /* me26_fsync */
    NULL,          /* me26_fasync */
                   /* nothing more, fill with NULLs */
};


int init_module(void)
{
  int flag_ok=0;
  int result;
  int flag_manual=0;
  /* Arrays for temporary storage of good results in probing */
  int me26_base_dt[ME26_NR_CARDS + 1];
  int me26_irq_dt[ME26_NR_CARDS];
  int dcn,i;
  unsigned data;
  Me26_Dev *dev; /* device information */

  /* Array of automatic configuration base addresses */
  int me26_base_d[] = {ME26_BASE1,ME26_BASE2,ME26_BASE3,ME26_BASE4,0};

  PDEBUG("Entering me26\n");    

  /* Initialization of base and irq's */
  me26_base_t[0]=me26_base;
  me26_base_t[1]=me26_base2;
  me26_base_t[2]=me26_base3;
  me26_base_t[3]=me26_base4;

  me26_irq_t[0]=me26_irq;
  me26_irq_t[1]=me26_irq2;
  me26_irq_t[2]=me26_irq3;
  me26_irq_t[3]=me26_irq4;

  /* We see if base address is manual or automatic */
  me26_base_d[me26_nr_cards]=0;
  for (me26_cn=0; me26_base_d[me26_cn]; me26_cn++ ) {
    if ( me26_base_t[me26_cn] == 0 ) {
      /* Automatic probing */
      me26_base_t[me26_cn] = me26_base_d[me26_cn];
    } else {
      /* Manual configuration, leave as is */
      flag_manual=1;
      break;
    }
  }

  /* Checking cards */
  dcn=0;
  for (me26_cn=0;me26_base_t[me26_cn]; me26_cn++) {
    PDEBUG( "me26[%d]: probing 0x%x\n",me26_cn, me26_base_t[me26_cn]);
    /* Registering IO/ports */
    result = check_region(me26_base_t[me26_cn],8);

    if (result) {
        printk(KERN_INFO "me26[%d]: can't get I/O address 0x%x\n",
	       me26_cn,me26_base_t[me26_cn]);
	continue;
    }
    request_region(me26_base_t[me26_cn],8,"me26");

    /* We set the board, the 8255 PPI */
    outb_p(0x92,me26_base_t[me26_cn] + CARD_CONTROL);
    data=inb_p(me26_base_t[me26_cn] + CARD_CONTROL);
    PDEBUG( "card  0x%x \n",data);

    /* Probing IRQ */
    me26_selfprobe(me26_cn);

    if ( me26_irq_t[me26_cn] <= 0 ) {
      /* IRQ not found */
      release_region(me26_base_t[me26_cn],8);
    } else {
      /* IRQ found */
      flag_ok=1;
      /* Save base and IRQ */
      me26_base_dt[dcn]=me26_base_t[me26_cn];
      me26_irq_dt[dcn]=me26_irq_t[me26_cn];
      dcn++;
    }

  }
  me26_base_dt[dcn]=0;

  for (dcn=0;me26_base_dt[dcn]; dcn++) {
    me26_base_t[dcn]=me26_base_dt[dcn];
    me26_irq_t[dcn]=me26_irq_dt[dcn];
    PDEBUG("me26[%d]: base 0x%x irq %d \n",dcn,
	   me26_base_t[dcn],me26_irq_t[dcn]);
  }
  me26_base_t[dcn]=0;

  me26_nr_cards=dcn;
  PDEBUG("Number of cards = %d\n",me26_nr_cards);


  /* No cards found */
  if (flag_manual == 1) {
    if (dcn != me26_cn) {
      result=-EBUSY;
      goto fail_register;
    }
  } else {
    if ( flag_ok == 0 )
      return -EBUSY;
  }

  /* Creating the inverse IRQ matrix */
  for (i=0;i<16;i++) {
    for (dcn=0;me26_base_dt[dcn]; dcn++) {
      if ( i == me26_irq_t[dcn]) {
	me26_card_t[i]=dcn;
	break;
      }
    }
  }

  /* Registering device numbers */
  result = register_chrdev(me26_major, "me26", &me26_fops);
  if (result < 0) {
    printk(KERN_INFO "me26: can't get major number\n");
    goto fail_register;
  }
  if (me26_major == 0) me26_major = result; /* dynamic */

  me26_devices=kmalloc(me26_nr_cards*sizeof(Me26_Dev),GFP_KERNEL);
  if (!me26_devices) {
    result=-ENOMEM;
    goto fail_malloc;
  }
  
  /* Setting buffer pointing to zero, requested int to 0,
   wait queues to NULL and dev->single to 0 */
  for (dcn=0;me26_base_dt[dcn]; dcn++) { /* Different card number */
    dev=&me26_devices[dcn];
    dev->req_int=0;
    me26_queue[dcn]=NULL;
    for (i=0;i<16;i++) { /* Different channel number */
      dev->buffer[i]=0;
      dev->single[i]=0;
    }
    dev->timbuf=0;
    dev->timsin=0;
  }

  return 0; /* succeed */

  fail_malloc: unregister_chrdev(me26_major,"me26");
  fail_register: for (me26_cn=0;me26_base_t[me26_cn]; me26_cn++) 
    release_region(me26_base_t[me26_cn],8);
  
  return result;


}

void cleanup_module(void)
{
        
  PDEBUG("Leaving me26\n");    
  unregister_chrdev(me26_major, "me26");
  for (me26_cn=0;me26_base_t[me26_cn]; me26_cn++) {
    release_region(me26_base_t[me26_cn],8);
  }
  kfree(me26_devices);
}
